#include <Arduino.h>
#include <Thread.h> // https://github.com/ivanseidel/ArduinoThread
#include <ThreadController.h>
#include <AsyncMqttClient.h>
// #include <AsyncUDP.h>
#include <Ticker.h>
#include "Streaming.h"
#include "main.h"
#include <AsyncTCP.h>
#include "ESPAsyncWebServer.h"
#include <ArduinoJson.h>
#include <Wire.h>
#include "MyTools.h"
#include <AsyncElegantOTA.h>
#include "mbedtls/aes.h"
#include "FileOperations.h"

// #define ETH_MODE

#ifndef ETH_MODE
#include <WiFi.h>
#include <WiFiAP.h>
#endif

char *aes_key = "xX17y#y%qhvMjt&9";

const char *host = "rfid_writer";
const char *version = "V_1.1.3";
// const char *cl_ssid = "biodit_sys";
// const char *cl_pass = "4ETj5dPy";
// const char *mqtt_server = "192.168.1.15";
uint64_t awake_timer = 0; // awake state
long factory_counter = 0;
#define FACTORY_RESET_TIME 5000 // 5 sec

IPAddress local_IP(192, 168, 1, 14);
// Set your Gateway IP address
IPAddress gateway(192, 168, 1, 1);
IPAddress subnet(255, 255, 255, 0);
IPAddress primaryDNS(8, 8, 8, 8);   // optional
IPAddress secondaryDNS(8, 8, 4, 4); // optional

FileOperations filesys;

String topic_late;
String message_late;

// --- RFID
#include <SPI.h>
#include <MFRC522.h>
#define RST_PIN 14                // Configurable, see typical pin layout above
#define SS_PIN 5                  // Configurable, see typical pin layout above
#define IRQ_PIN 2                 // Configurable, depends on hardware
MFRC522 mfrc522(SS_PIN, RST_PIN); // Create MFRC522 instance.
MFRC522::MIFARE_Key key;
MFRC522::StatusCode status;
// Init array that will store new NUID
byte nuidPICC[4];
byte card_block1[18];
byte card_block2[18];
byte card_block4[18];
byte card_block5[18];
byte card_block6[18];

WiFiClient espClient;
AsyncMqttClient mqttClient;
Ticker mqttReconnectTimer;
IPAddress IP;

#define DEVICE_TYPE "writer"
#define LED_RED 4
#define LED_GREEN 21
#define ZUMER 22
#define ENABLE_PERIPHERY 25 // Olock compability
#define WIFI_TIME 3500      // 5 sec
#define CARD_TIMEOUT 5000
#define FACTORY_RESET 39 // needs to be defined
#define CONNFIG_FILE "/configjson.txt"

char MAC_char[18];
uint8_t MAC_array[6];
const char *mymac = "240ac40457d0";
const char *type = DEVICE_TYPE;
char clientName[52] = "Olock_xxxx";
char out_message_buffer[2048];
char operations_buffer[2048];
uint32_t startSound = 0;
AsyncWebServer server(80);
MyTools tools;
bool factory_reset_state = false;
bool factory_mode = false;

// CONFIG PARAMETERS
const char *PARAM_HOTEL = "hotel";
const char *PARAM_NAME = "name";
const char *PARAM_SSID = "ssid";
const char *PARAM_PASS = "pass";
const char *PARAM_MQTT = "mqtt";
const char *PARAM_MQTT_PORT = "mqtt_port";
const char *PARAM_MQTT_USER = "mqtt_user";
const char *PARAM_MQTT_USER_PASS = "mqtt_user_pass";
const char *PARAM_IP = "ip_addres";
const char *PARAM_MASK = "mask";
const char *PARAM_GATEWAY = "gateway";
const char *PARAM_DHCP = "dhcp";
const char *PARAM_TYPE = "type";
const char *PARAM_SOUND = "sound";
const char *PARAM_GROUP = "group";

struct Config
{
  uint16_t hotel_code;
  uint16_t room;
  uint16_t group;
  char name[16];
  char ssid[30];
  char pass[30];
  char mqtt[30];
  char mqtt_user[16];
  char mqtt_user_pass[16];
  int mqtt_port;
  int dhcp;
  char ip_address[16];
  char ip_mask[16];
  char gateway[16];
  char type[16];
  int relay_time = 3;
  char event[16];
  int sound;
  char apssid[16];
  char appass[16];
  char update_code[16]; // code for activating update card
  uint8_t today;        // getting time once a day after event or timer wakeup
};
RTC_DATA_ATTR Config config;

/*
 ######  ######## ######## ##     ## ########
##    ## ##          ##    ##     ## ##     ##
##       ##          ##    ##     ## ##     ##
 ######  ######      ##    ##     ## ########
      ## ##          ##    ##     ## ##
##    ## ##          ##    ##     ## ##
 ######  ########    ##     #######  ##
*/
void setup()
{
  Serial.begin(115200);
  // SET PINS

  initGPIO();
  digitalWrite(ZUMER, 0); // stop sound
                          // FACTORY_RESET
  if (digitalRead(FACTORY_RESET) == 0)
  {
    factory_reset_state = true;
    activate_AP();
    return;
  }

  // File System
  if (!SPIFFS.begin())
  {
    Serial.println(F("Card Mount Failed"));
    errorSound();
    esp_restart();
  }

  loadConfiguration();
  Serial << "loaded configuration" << endl;
  Serial << "DHCP:" << config.dhcp << endl;
  // Init buffer
  for (byte i = 0; i < 6; i++)
  {
    key.keyByte[i] = 0xFF;
  }
  initSPI();
  connectWiFi();
  Serial << "Connecting to MQTT: " << config.mqtt << endl;
  Serial << "Connecting to MQTT PORT: " << config.mqtt_port << endl;
  connectBroker();
}

/*
##     ##    ###    #### ##    ##    ##        #######   #######  ########
###   ###   ## ##    ##  ###   ##    ##       ##     ## ##     ## ##     ##
#### ####  ##   ##   ##  ####  ##    ##       ##     ## ##     ## ##     ##
## ### ## ##     ##  ##  ## ## ##    ##       ##     ## ##     ## ########
##     ## #########  ##  ##  ####    ##       ##     ## ##     ## ##
##     ## ##     ##  ##  ##   ###    ##       ##     ## ##     ## ##
##     ## ##     ## #### ##    ##    ########  #######   #######  ##
*/
void loop()
{
  if (!factory_reset_state)
  {
    // readCard();
    if (digitalRead(FACTORY_RESET))
    {
      // not pressed
      digitalWrite(LED_RED, 0);
      factory_counter = millis();
    }
    else
    {
      digitalWrite(LED_RED, 1);
      if (millis() > factory_counter + FACTORY_RESET_TIME)
      {
        factory_reset_state = true;
        factory_counter = millis();
        activate_AP();
      }
    }
    messageLoop();
  }
  soundLoop();
  delay(20);
}

/**
 * @brief Stops sound after given period
 *
 */
void soundLoop()
{
  if (startSound > millis())
    return;
  // timer
  digitalWrite(ZUMER, 0);
}

/**
 * @brief Set the Sound in milliseconds
 *
 * @param time
 */
void setSound(uint16_t time)
{
  digitalWrite(ZUMER, 1);
  startSound = millis() + time;
}

/**
 * @brief Message loop
 *
 */
void messageLoop()
{
  if (topic_late.length() == 0 || message_late.length() == 0)
    return;
  mac_subscriber(topic_late, message_late);
  topic_late.clear();
  message_late.clear();
}

//  ######   ########  ####  #######
// ##    ##  ##     ##  ##  ##     ##
// ##        ##     ##  ##  ##     ##
// ##   #### ########   ##  ##     ##
// ##    ##  ##         ##  ##     ##
// ##    ##  ##         ##  ##     ##
//  ######   ##        ####  #######

void initGPIO()
{
  pinMode(FACTORY_RESET, INPUT);
  pinMode(ZUMER, OUTPUT);
  pinMode(ENABLE_PERIPHERY, OUTPUT);
  pinMode(LED_RED, OUTPUT);
  pinMode(LED_GREEN, OUTPUT);
  digitalWrite(ENABLE_PERIPHERY, HIGH); // enable periphery
}

void initSPI()
{
  // Init SPI bus
  SPI.begin();
  mfrc522.PCD_Init(); // Init MFRC522 card
  mfrc522.PCD_SoftPowerUp();
}

// ########  ########   #######  ##    ## ######## ########
// ##     ## ##     ## ##     ## ##   ##  ##       ##     ##
// ##     ## ##     ## ##     ## ##  ##   ##       ##     ##
// ########  ########  ##     ## #####    ######   ########
// ##     ## ##   ##   ##     ## ##  ##   ##       ##   ##
// ##     ## ##    ##  ##     ## ##   ##  ##       ##    ##
// ########  ##     ##  #######  ##    ## ######## ##     ##

/**
 * @brief Connect to MQTT brocker
 *
 */
void connectBroker()
{
  // uint8_t tries = 5;
  // while (!mqttClient.connected())
  // {
  mqttClient.onConnect(onMqttConnect);
  mqttClient.onDisconnect(onMqttDisconnect);
  mqttClient.onSubscribe(onMqttSubscribe);
  mqttClient.onUnsubscribe(onMqttUnsubscribe);
  mqttClient.onPublish(onMqttPublish);
  // mqttClient.setServer(MQTT_HOST, MQTT_PORT);

  mqttClient.setCleanSession(true);
  mqttClient.setClientId(clientName);
  mqttClient.setServer(config.mqtt, config.mqtt_port);
  char will[30];
  sprintf(will, "writer/disconnect/%s", mymac);
  mqttClient.setWill("system", 0, false, will, 30);
  mqttClient.onMessage(onMqttMessage);
  mqttClient.connect();
  delay(200);
  Serial << "." << endl;
  awake_timer = millis();
  while (!mqttClient.subscribe(mymac, 2))
  {
    Serial << "sub ..." << endl;
    digitalWrite(LED_RED, 1); // blink on connection tryes
    delay(100);
    digitalWrite(LED_RED, 0);
    delay(100);
    if (millis() - awake_timer > WIFI_TIME)
    {
      esp_restart();
    }
  };
}

void onMqttConnect(bool sessionPresent)
{
  // mqttClient.subscribe("all", 0);
  Serial.println("Connected to MQTT.");
  responseAlive();
}

void onMqttPublish(uint16_t packetId)
{
  Serial.println("Publish acknowledged.");
  Serial.print("  packetId: ");
  Serial.println(packetId);
}
void onMqttUnsubscribe(uint16_t packetId)
{
  Serial.println("Unsubscribe acknowledged.");
  Serial.print("  packetId: ");
  Serial.println(packetId);
}
void connectToMqtt()
{
  Serial.println("Connecting to MQTT...");
  connectBroker();
}

void onMqttDisconnect(AsyncMqttClientDisconnectReason reason)
{
  Serial.println("Disconnected from MQTT.");

  if (reason == AsyncMqttClientDisconnectReason::TLS_BAD_FINGERPRINT)
  {
    Serial.println("Bad server fingerprint.");
  }

  if (WiFi.isConnected())
  {
    mqttReconnectTimer.once(2, connectToMqtt);
  }
}

/**
 * @brief On subscription
 *
 * @param packetId
 * @param qos
 */
void onMqttSubscribe(uint16_t packetId, uint8_t qos)
{
  Serial.println(F("Subscribe acknowledged."));
  Serial.print(F("  packetId: "));
  Serial.println(packetId);
  Serial.print(F("  qos: "));
  Serial.println(qos);
}

/**
 * @brief  MQTT message receiver
 *
 * @param topic
 * @param payload
 * @param properties
 * @param len
 * @param index
 * @param total
 */
void onMqttMessage(char *topic, char *payload, AsyncMqttClientMessageProperties properties, size_t len, size_t index, size_t total)
{
  Serial.print(F("Received: "));
  // mqttClient.disconnect();
  // mqttClient.unsubscribe(mymac);
  String message = "";
  for (size_t i = 0; i < len; i++)
  {
    message += payload[i];
  }
  Serial << topic << ":" << message << endl;
  topic_late = topic;
  message_late = message;
}

// ##      ## #### ######## ####
// ##  ##  ##  ##  ##        ##
// ##  ##  ##  ##  ##        ##
// ##  ##  ##  ##  ######    ##
// ##  ##  ##  ##  ##        ##
// ##  ##  ##  ##  ##        ##
//  ###  ###  #### ##       ####

void connectWiFi()
{
  // Connect to WiFi
  Serial << "Connecting to WiFi: " << config.ssid << endl;
  WiFi.disconnect();
  char host[16];
  sprintf(host, "writer_%d", config.room);
  WiFi.setHostname(host);
  if (config.dhcp == 1)
  {
    Serial << "using DHCP" << endl;
    WiFi.begin(config.ssid, config.pass);
  }
  else
  {
    local_IP.fromString(config.ip_address);
    gateway.fromString(config.gateway);
    subnet.fromString(config.ip_mask);
    if (!WiFi.config(local_IP, gateway, subnet, primaryDNS, secondaryDNS))
    {
      Serial.println("STA Failed to configure");
    }
    else
    {
      Serial << "Static IP" << endl;
      Serial << local_IP << ", " << gateway << ", " << subnet << ", " << primaryDNS << ", " << secondaryDNS << endl;
      Serial << WiFi.localIP() << ", " << WiFi.gatewayIP() << ", " << WiFi.subnetMask() << ", " << WiFi.dnsIP() << endl;
      WiFi.begin(config.ssid, config.pass);
    }
  }

  //---------------
  awake_timer = millis();
  while (WiFi.status() != WL_CONNECTED)
  {
    // TODO set timeout to go sleep
    delay(100);
    digitalWrite(ZUMER, 0); // stop sound
    Serial.print(".");
    // restarts if no Wifi connection
    if (millis() - awake_timer > WIFI_TIME)
    {
      esp_restart();
    }
  }

  Serial.println(F("connected"));
  // GET MAC
  WiFi.macAddress(MAC_array);
  for (int i = 0; i < sizeof(MAC_array); ++i)
  {
    sprintf(MAC_char, "%s%02x", MAC_char, MAC_array[i]);
  }

  mymac = MAC_char;
  // Set client name
  // Writer_240ac40457d0
  sprintf(clientName, "Writer_%s", mymac);
  Serial << "MQTT client name:" << clientName << endl;
  IP = WiFi.localIP();
}

// ##     ##  #######  ######## ########
// ###   ### ##     ##    ##       ##
// #### #### ##     ##    ##       ##
// ## ### ## ##     ##    ##       ##
// ##     ## ##  ## ##    ##       ##
// ##     ## ##    ##     ##       ##
// ##     ##  ##### ##    ##       ##

void confirmOkReceiving()
{
  String card_num;
  card_num += tools.byteToHex(mfrc522.uid.uidByte[0]);
  card_num += tools.byteToHex(mfrc522.uid.uidByte[1]);
  card_num += tools.byteToHex(mfrc522.uid.uidByte[2]);
  card_num += tools.byteToHex(mfrc522.uid.uidByte[3]);
  uint32_t cardint = 0;
  cardint += mfrc522.uid.uidByte[3] * 16777216;
  cardint += mfrc522.uid.uidByte[2] * 65536;
  cardint += mfrc522.uid.uidByte[1] * 256;
  cardint += mfrc522.uid.uidByte[0];
  //[MAC]/OK
  if (!mqttClient.connected())
    connectBroker();
  sprintf(out_message_buffer, "%s/OK/%s/%010d", mymac, card_num.c_str(), cardint);
  Serial << "Confirm message:" << out_message_buffer << endl;
  mqttClient.publish("confirm", 0, false, out_message_buffer);
}

void confirmNokReceiving(const char *message)
{
  //[MAC]/NOK/[message]
  if (!mqttClient.connected())
    connectBroker();
  sprintf(out_message_buffer, "%s/NOK/%s", mymac, message);
  Serial << "Confirm message:" << out_message_buffer << endl;
  mqttClient.publish("confirm", 0, false, out_message_buffer);
}

/**
 * @brief  RESPONSE ALIVE
 */
void responseAlive()
{
  String message(MAC_char);
  message += "/";
  message += type;
  message += "/";
  message += IP.toString().c_str();
  message += "/";
  message += version;
  message += "/"; // no name
  message += "/no battery";
  message += "/";
  message += WiFi.RSSI();
  message += "/";
  mqttClient.publish("alive", 1, false, message.c_str());
}

/**
 * @brief Response on topic own MAC
 *
 * @param topic
 * @param payload
 * @param length
 *
 */
void mac_subscriber(String topic, String payload)
{
  if (payload.equals("ping"))
  {
    blink();
    responseAlive();
    return;
  }
  if (payload.equals("read"))
  {
    readCardInfo();
  }
  if (payload.equals("read_raw"))
  {
    readCardRawInfo();
  }
  if (payload.equals("update"))
  {
    startServer();
  }
  if (payload.equals("test"))
  {
    testAES();
  }
  if (payload.equals("restart"))
  {
    esp_restart();
  }

  String msg[7] = {"", "", "", "", "", "", ""};
  int chunks = tools.parseMessage(msg, payload);
  Serial << "chunks:" << chunks << endl;
  // if (chunks < 3)
  //   return;

  if (msg[0].equals("write"))
  {
    Serial << "write card" << endl;
    uint8_t buff1[16];
    uint8_t buff2[16];
    uint8_t buff4[16];
    uint8_t buff5[16];
    uint8_t buff6[16];
    for (size_t i = 0; i < 16; i++)
    {
      buff1[i] = tools.hex2bin(msg[1].charAt(i * 2)) * 16 + tools.hex2bin(msg[1].charAt(i * 2 + 1));
      buff2[i] = tools.hex2bin(msg[2].charAt(i * 2)) * 16 + tools.hex2bin(msg[2].charAt(i * 2 + 1));
      if (!msg[3].equals(""))
      {
        buff4[i] = tools.hex2bin(msg[3].charAt(i * 2)) * 16 + tools.hex2bin(msg[3].charAt(i * 2 + 1));
      }
      if (!msg[4].equals(""))
      {
        buff5[i] = tools.hex2bin(msg[4].charAt(i * 2)) * 16 + tools.hex2bin(msg[4].charAt(i * 2 + 1));
      }
      if (!msg[5].equals(""))
      {
        buff6[i] = tools.hex2bin(msg[5].charAt(i * 2)) * 16 + tools.hex2bin(msg[5].charAt(i * 2 + 1));
      }
      // Serial<<msg[1].charAt(i * 2)<<msg[1].charAt(i * 2 + 1)<<"->"<<buff1[i]<<endl;
    }
    tools.dump_byte_array(buff1, 16);
    tools.dump_byte_array(buff2, 16);
    tools.dump_byte_array(buff4, 16);
    tools.dump_byte_array(buff5, 16);
    tools.dump_byte_array(buff6, 16);
    writeBlocks(buff1, buff2, buff4, buff5, buff6);
    mfrc522.PICC_HaltA();      // Halt PICC
    mfrc522.PCD_StopCrypto1(); // Stop encryption on PCD
  }
}

// ########  ########    ###    ########         ######     ###    ########  ########
// ##     ## ##         ## ##   ##     ##       ##    ##   ## ##   ##     ## ##     ##
// ##     ## ##        ##   ##  ##     ##       ##        ##   ##  ##     ## ##     ##
// ########  ######   ##     ## ##     ##       ##       ##     ## ########  ##     ##
// ##   ##   ##       ######### ##     ##       ##       ######### ##   ##   ##     ##
// ##    ##  ##       ##     ## ##     ##       ##    ## ##     ## ##    ##  ##     ##
// ##     ## ######## ##     ## ########         ######  ##     ## ##     ## ########

void readCard()
{
// Clear buffer
  for (size_t i = 0; i < 16; i++)
  {
    card_block1[i] = 0;
    card_block2[i] = 0;
    card_block4[i] = 0;
    card_block5[i] = 0;
    card_block6[i] = 0;
  }

  digitalWrite(ZUMER, 1);
  digitalWrite(LED_RED, 1);
  digitalWrite(LED_GREEN, 1);
  delay(100);
  digitalWrite(ZUMER, 0);
  int time = millis();
  mfrc522.PCD_Init(); // Init MFRC522 card
  // mfrc522.PCD_SoftPowerUp();
  delay(100);
  while (!mfrc522.PICC_IsNewCardPresent())
  {
    if (millis() > time + CARD_TIMEOUT)
    {
      // wait 5 secconds
      digitalWrite(LED_GREEN, 0);
      errorSound();
      return;
    }
  }
  digitalWrite(LED_GREEN, 0);
  // Verify if the NUID has been readed
  if (!mfrc522.PICC_ReadCardSerial())
  {
    digitalWrite(LED_GREEN, 0);
    errorSound();
    return;
  }
  MFRC522::PICC_Type piccType = mfrc522.PICC_GetType(mfrc522.uid.sak);
  // Serial.println(mfrc522.PICC_GetTypeName(piccType));
  if (piccType != MFRC522::PICC_TYPE_MIFARE_1K && piccType != MFRC522::PICC_TYPE_MIFARE_4K)
  {
    Serial.println(F("This reader only works with MIFARE Classic cards."));
    errorSound();
    // Halt PICC
    mfrc522.PICC_HaltA();
    // Stop encryption on PCD
    mfrc522.PCD_StopCrypto1();
    digitalWrite(LED_GREEN, 0);
    return;
  }
  for (byte i = 0; i < 4; i++)
  {
    nuidPICC[i] = mfrc522.uid.uidByte[i];
  }
  digitalWrite(LED_GREEN, 0);
  digitalWrite(LED_RED, 0);
  readBlock(1, card_block1);
  readBlock(2, card_block2);
  readBlock(4, card_block4);
  readBlock(5, card_block5);
  readBlock(6, card_block6);
  delay(500);
  digitalWrite(LED_GREEN, 1);
  good();
  digitalWrite(LED_GREEN, 0);
  // Halt PICC
  mfrc522.PICC_HaltA();
  // Stop encryption on PCD
  mfrc522.PCD_StopCrypto1();
  digitalWrite(LED_GREEN, 0);
}
void readCardRawInfo()
{
  readCard();
  publishRawCardInfo();
}



/**
 * @brief Read card info
 *
 */
void readCardInfo()
{
  readCard();
  publishLastCardInfo();
}

/**
 * @brief Read card blocks
 *
 * @param block
 * @param buffer
 */
void readBlock(byte block, byte *buffer)
{
  byte len = 18;
  MFRC522::StatusCode status;
  // byte buffer1[18];
  // Serial<<"block: "<<block<<endl;
  status = mfrc522.PCD_Authenticate(MFRC522::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid)); // line 834 of MFRC522.cpp file
  if (status != MFRC522::STATUS_OK)
  {
    Serial.print(F("Authentication failed: "));
    Serial.println(mfrc522.GetStatusCodeName(status));
    return;
  }

  status = mfrc522.MIFARE_Read(block, buffer, &len);
  if (status != MFRC522::STATUS_OK)
  {
    Serial.print(F("Reading failed: "));
    Serial.println(mfrc522.GetStatusCodeName(status));
    return;
  }
}


void publishRawCardInfo()
{
if (!mqttClient.connected())
  {
    connectBroker();
  }
  String card_num;
  card_num += tools.byteToHex(nuidPICC[0]);
  card_num += tools.byteToHex(nuidPICC[1]);
  card_num += tools.byteToHex(nuidPICC[2]);
  card_num += tools.byteToHex(nuidPICC[3]);
  uint32_t cardint = 0;
  cardint += mfrc522.uid.uidByte[3] * 16777216;
  cardint += mfrc522.uid.uidByte[2] * 65536;
  cardint += mfrc522.uid.uidByte[1] * 256;
  cardint += mfrc522.uid.uidByte[0];

  char card[9];
  card_num.toCharArray(card, 9);

  String buf1 = "";
  String buf2 = "";
  String buf4 = "";
  String buf5 = "";
  String buf6 = "";
  for (uint8_t i = 0; i < 16; i++)
  {
    buf1 += tools.byteToHex(card_block1[i]);
    buf2 += tools.byteToHex(card_block2[i]);
    buf4 += tools.byteToHex(card_block4[i]);
    buf5 += tools.byteToHex(card_block5[i]);
    buf6 += tools.byteToHex(card_block6[i]);
  }
  char bf1[33];
  char bf2[33];
  char bf4[33];
  char bf5[33];
  char bf6[33];

  buf1.toCharArray(bf1, 33);
  buf2.toCharArray(bf2, 33);
  buf4.toCharArray(bf4, 33);
  buf5.toCharArray(bf5, 33);
  buf6.toCharArray(bf6, 33);

  // sprintf(out_message_buffer, "%s/%s/%s/%s", mymac, card, bf1, bf2);
  sprintf(out_message_buffer, "%s/%s/%s/%s/%s/%s/%s/%010d", mymac, card, bf1, bf2, bf4, bf5, bf6, cardint);
  Serial << "Send message:" << out_message_buffer << endl;
  mqttClient.publish("raw_card", 0, false, out_message_buffer);
}

/**
 * @brief Publish card info to broker
 *
 */
void publishLastCardInfo()
{
  // on topic "card" - message: [MAC]/access/[slot]/[user_id]/[time]/[shift]
  if (!mqttClient.connected())
  {
    connectBroker();
  }
  String card_num;
  card_num += tools.byteToHex(nuidPICC[0]);
  card_num += tools.byteToHex(nuidPICC[1]);
  card_num += tools.byteToHex(nuidPICC[2]);
  card_num += tools.byteToHex(nuidPICC[3]);
  uint32_t cardint = 0;
  cardint += mfrc522.uid.uidByte[3] * 16777216;
  cardint += mfrc522.uid.uidByte[2] * 65536;
  cardint += mfrc522.uid.uidByte[1] * 256;
  cardint += mfrc522.uid.uidByte[0];

  char card[9];
  card_num.toCharArray(card, 9);

  String buf1 = "";
  String buf2 = "";
  String buf4 = "";
  String buf5 = "";
  String buf6 = "";
  // for (uint8_t i = 0; i < 16; i++)
  // {
  //   buf1 += tools.byteToHex(card_block1[i]);
  //   buf2 += tools.byteToHex(card_block2[i]);
  //   buf4 += tools.byteToHex(card_block4[i]);
  //   buf5 += tools.byteToHex(card_block5[i]);
  //   buf6 += tools.byteToHex(card_block6[i]);
  // }
  char bf1[33];
  char bf2[33];
  char bf4[33];
  char bf5[33];
  char bf6[33];

  decrypt((char *)card_block1, aes_key, (unsigned char *)bf1);
  for (size_t i = 0; i < 16; i++)
  {
    buf1 += tools.byteToHex(bf1[i]);
  }
  decrypt((char *)card_block2, aes_key, (unsigned char *)bf2);
  for (size_t i = 0; i < 16; i++)
  {
    buf2 += tools.byteToHex(bf2[i]);
  }
  decrypt((char *)card_block4, aes_key, (unsigned char *)bf4);
  for (size_t i = 0; i < 16; i++)
  {
    buf4 += tools.byteToHex(bf4[i]);
  }
  decrypt((char *)card_block5, aes_key, (unsigned char *)bf5);
  for (size_t i = 0; i < 16; i++)
  {
    buf5 += tools.byteToHex(bf5[i]);
  }
  decrypt((char *)card_block6, aes_key, (unsigned char *)bf6);
  for (size_t i = 0; i < 16; i++)
  {
    buf6 += tools.byteToHex(bf6[i]);
  }

  buf1.toCharArray(bf1, 33);
  buf2.toCharArray(bf2, 33);
  buf4.toCharArray(bf4, 33);
  buf5.toCharArray(bf5, 33);
  buf6.toCharArray(bf6, 33);

  // sprintf(out_message_buffer, "%s/%s/%s/%s", mymac, card, bf1, bf2);
  sprintf(out_message_buffer, "%s/%s/%s/%s/%s/%s/%s/%010d", mymac, card, bf1, bf2, bf4, bf5, bf6, cardint);
  Serial << "Send message:" << out_message_buffer << endl;
  mqttClient.publish("card", 0, false, out_message_buffer);
}

// ##      ## ########  #### ######## ########        ######     ###    ########  ########
// ##  ##  ## ##     ##  ##     ##    ##             ##    ##   ## ##   ##     ## ##     ##
// ##  ##  ## ##     ##  ##     ##    ##             ##        ##   ##  ##     ## ##     ##
// ##  ##  ## ########   ##     ##    ######         ##       ##     ## ########  ##     ##
// ##  ##  ## ##   ##    ##     ##    ##             ##       ######### ##   ##   ##     ##
// ##  ##  ## ##    ##   ##     ##    ##             ##    ## ##     ## ##    ##  ##     ##
//  ###  ###  ##     ## ####    ##    ########        ######  ##     ## ##     ## ########

void writeBlocks(byte *buffer1, byte *buffer2, byte *buffer4, byte *buffer5, byte *buffer6)
{
  mfrc522.PCD_Init(); // Init MFRC522 card
                      // mfrc522.PCD_SoftPowerUp();
  delay(100);
  // Prepare key - all keys are set to FFFFFFFFFFFFh at chip delivery from the factory.
  MFRC522::MIFARE_Key key;
  for (byte i = 0; i < 6; i++)
    key.keyByte[i] = 0xFF;

  // Reset the loop if no new card present on the sensor/reader. This saves the entire process when idle.
  digitalWrite(ZUMER, 1);
  digitalWrite(LED_GREEN, 1);
  delay(100);
  digitalWrite(ZUMER, 0);
  // digitalWrite(LED_GREEN, 0);
  digitalWrite(LED_RED, 1);
  int time = millis();
  while (!mfrc522.PICC_IsNewCardPresent())
  {
    if (millis() > time + CARD_TIMEOUT) // 5 secconds
      break;
  }
  // Select one of the cards
  if (!mfrc522.PICC_ReadCardSerial())
  {
    Serial << "No card" << endl;
    errorSound();
    confirmNokReceiving("No card found!");
    digitalWrite(LED_GREEN, 0);
    return;
  }

  MFRC522::StatusCode status;

  byte block = 1;
  // Serial.println(F("Authenticating using key A..."));
  status = mfrc522.PCD_Authenticate(MFRC522::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522::STATUS_OK)
  {
    Serial.print(F("PCD_Authenticate() failed: "));
    Serial.println(mfrc522.GetStatusCodeName(status));
    errorSound();
    confirmNokReceiving("Can't authenticate!");
    digitalWrite(LED_GREEN, 0);
    return;
  }

  // Write block
  status = mfrc522.MIFARE_Write(block, buffer1, 16);
  if (status != MFRC522::STATUS_OK)
  {
    Serial.print(F("MIFARE_Write() failed: "));
    Serial.println(mfrc522.GetStatusCodeName(status));
    errorSound();
    confirmNokReceiving("Writing failed!");
    digitalWrite(LED_GREEN, 0);
    return;
  }

  block = 2;
  // Serial.println(F("Authenticating using key A..."));
  status = mfrc522.PCD_Authenticate(MFRC522::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522::STATUS_OK)
  {
    Serial.print(F("PCD_Authenticate() failed: "));
    Serial.println(mfrc522.GetStatusCodeName(status));
    errorSound();
    confirmNokReceiving("Can't authenticate!");
    digitalWrite(LED_GREEN, 0);
    return;
  }

  // Write block
  status = mfrc522.MIFARE_Write(block, buffer2, 16);
  if (status != MFRC522::STATUS_OK)
  {
    Serial.print(F("MIFARE_Write() failed: "));
    Serial.println(mfrc522.GetStatusCodeName(status));
    errorSound();
    confirmNokReceiving("Writing failed!");
    digitalWrite(LED_GREEN, 0);
    return;
  }

  block = 4;
  // Serial.println(F("Authenticating using key A..."));
  status = mfrc522.PCD_Authenticate(MFRC522::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522::STATUS_OK)
  {
    Serial.print(F("PCD_Authenticate() failed: "));
    Serial.println(mfrc522.GetStatusCodeName(status));
    errorSound();
    confirmNokReceiving("Can't authenticate!");
    digitalWrite(LED_GREEN, 0);
    return;
  }

  // Write block
  status = mfrc522.MIFARE_Write(block, buffer4, 16);
  if (status != MFRC522::STATUS_OK)
  {
    Serial.print(F("MIFARE_Write() failed: "));
    Serial.println(mfrc522.GetStatusCodeName(status));
    errorSound();
    confirmNokReceiving("Writing failed!");
    digitalWrite(LED_GREEN, 0);
    return;
  }

  block = 5;
  // Serial.println(F("Authenticating using key A..."));
  status = mfrc522.PCD_Authenticate(MFRC522::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522::STATUS_OK)
  {
    Serial.print(F("PCD_Authenticate() failed: "));
    Serial.println(mfrc522.GetStatusCodeName(status));
    errorSound();
    confirmNokReceiving("Can't authenticate!");
    digitalWrite(LED_GREEN, 0);
    return;
  }

  // Write block
  status = mfrc522.MIFARE_Write(block, buffer5, 16);
  if (status != MFRC522::STATUS_OK)
  {
    Serial.print(F("MIFARE_Write() failed: "));
    Serial.println(mfrc522.GetStatusCodeName(status));
    errorSound();
    confirmNokReceiving("Writing failed!");
    digitalWrite(LED_GREEN, 0);
    return;
  }

  block = 6;
  // Serial.println(F("Authenticating using key A..."));
  status = mfrc522.PCD_Authenticate(MFRC522::PICC_CMD_MF_AUTH_KEY_A, block, &key, &(mfrc522.uid));
  if (status != MFRC522::STATUS_OK)
  {
    Serial.print(F("PCD_Authenticate() failed: "));
    Serial.println(mfrc522.GetStatusCodeName(status));
    errorSound();
    confirmNokReceiving("Can't authenticate!");
    digitalWrite(LED_GREEN, 0);
  }

  // Write block
  status = mfrc522.MIFARE_Write(block, buffer6, 16);
  if (status != MFRC522::STATUS_OK)
  {
    Serial.print(F("MIFARE_Write() failed: "));
    Serial.println(mfrc522.GetStatusCodeName(status));
    errorSound();
    confirmNokReceiving("Writing failed!");
    digitalWrite(LED_GREEN, 0);
    return;
  }
  digitalWrite(LED_GREEN, 1);
  digitalWrite(LED_RED, 0);
  good();
  digitalWrite(LED_GREEN, 0);
  confirmOkReceiving();
  digitalWrite(ZUMER, 0);
  digitalWrite(LED_RED, 0);
  return;
  // mfrc522.PICC_HaltA();      // Halt PICC
  // mfrc522.PCD_StopCrypto1(); // Stop encryption on PCD
}

void encrypt(char *plainText, const char *key, unsigned char *outputBuffer)
{

  mbedtls_aes_context aes;

  mbedtls_aes_init(&aes);
  mbedtls_aes_setkey_enc(&aes, (const unsigned char *)key, strlen(key) * 8);
  mbedtls_aes_crypt_ecb(&aes, MBEDTLS_AES_ENCRYPT, (const unsigned char *)plainText, outputBuffer);
  mbedtls_aes_free(&aes);
}

void decrypt(char *chipherText, const char *key, unsigned char *outputBuffer)
{

  mbedtls_aes_context aes;

  mbedtls_aes_init(&aes);
  mbedtls_aes_setkey_dec(&aes, (const unsigned char *)key, strlen(key) * 8);
  mbedtls_aes_crypt_ecb(&aes, MBEDTLS_AES_DECRYPT, (const unsigned char *)chipherText, outputBuffer);
  mbedtls_aes_free(&aes);
}

void testCardDecode()
{
  // char *input = "Tech tutorials x";
  unsigned char output[16];
  decrypt((char *)card_block1, aes_key, output);
  String buff;
  for (size_t i = 0; i < 16; i++)
  {
    buff += tools.byteToHex(output[i]);
    Serial.print(tools.byteToHex(output[i]));
  }
  Serial.println();
  buff += "/";
  decrypt((char *)card_block2, aes_key, output);
  for (size_t i = 0; i < 16; i++)
  {
    buff += tools.byteToHex(output[i]);
    Serial.print(tools.byteToHex(output[i]));
  }
  Serial.println();
  Serial.println("key:");
  for (size_t j = 0; j < 16; j++)
  {
    Serial.print(tools.byteToHex(aes_key[j]));
  }
  Serial.println();

  // unsigned char decrypted[16];
  // decrypt(output,aes_key,decrypted);
  // for (size_t i = 0; i < 16; i++)
  // {
  //   /* code */
  //   Serial.print(tools.byteToHex(decrypted[i]));
  // }
  // Serial.println();
}

//  ######   #######  ##     ## ##    ## ########
// ##    ## ##     ## ##     ## ###   ## ##     ##
// ##       ##     ## ##     ## ####  ## ##     ##
//  ######  ##     ## ##     ## ## ## ## ##     ##
//       ## ##     ## ##     ## ##  #### ##     ##
// ##    ## ##     ## ##     ## ##   ### ##     ##
//  ######   #######   #######  ##    ## ########

/**
 * @brief Indicate errors
 *
 */
void errorSound()
{
  for (uint8_t i = 0; i < 5; i++)
  {
    digitalWrite(ZUMER, 1);
    digitalWrite(LED_RED, 1);
    delay(50);
    digitalWrite(ZUMER, 0);
    digitalWrite(LED_RED, 0);
    delay(50);
  }
}

void good()
{
  digitalWrite(ZUMER, 1);
  digitalWrite(LED_GREEN, 1);
  delay(500);
  digitalWrite(ZUMER, 0);
  digitalWrite(LED_GREEN, 0);
}

void encrypt(char *plainText, char *key, unsigned char *outputBuffer)
{
  mbedtls_aes_context aes;
  mbedtls_aes_init(&aes);
  mbedtls_aes_setkey_enc(&aes, (const unsigned char *)key, strlen(key) * 8);
  mbedtls_aes_crypt_ecb(&aes, MBEDTLS_AES_ENCRYPT, (const unsigned char *)plainText, outputBuffer);
  mbedtls_aes_free(&aes);
}

void decrypt(unsigned char *chipherText, char *key, unsigned char *outputBuffer)
{
  mbedtls_aes_context aes;
  mbedtls_aes_init(&aes);
  mbedtls_aes_setkey_dec(&aes, (const unsigned char *)key, strlen(key) * 8);
  mbedtls_aes_crypt_ecb(&aes, MBEDTLS_AES_DECRYPT, (const unsigned char *)chipherText, outputBuffer);
  mbedtls_aes_free(&aes);
}

void testAES()
{

  // TEST AES
  Serial.println("plain:");
  for (size_t i = 0; i < 16; i++)
  {
    /* code */
    Serial.print(tools.byteToHex(card_block1[i]));
  }
  Serial.println();
  Serial.println("key:");
  for (size_t j = 0; j < 16; j++)
  {
    Serial.print(tools.byteToHex(aes_key[j]));
  }
  Serial.println();
  // char *input = "Tech tutorials x";
  unsigned char output[128];
  encrypt((char *)card_block1, aes_key, output);
  for (size_t i = 0; i < 16; i++)
  {
    /* code */
    Serial.print(tools.byteToHex(output[i]));
  }
  Serial.println();
  // encrypt((char *)card_block2, aes_key, output);
  // for (size_t i = 0; i < 16; i++)
  // {
  //   /* code */
  //   Serial.print(tools.byteToHex(output[i]));
  // }
  // Serial.println();

  Serial.println(" decripted ");
  unsigned char output_buffer[16];
  decrypt(output, aes_key, output_buffer);
  for (size_t i = 0; i < 16; i++)
  {
    /* code */
    Serial.print(tools.byteToHex(output_buffer[i]));
  }
  Serial.println();
}

/*
   ###    ########
  ## ##   ##     ##
 ##   ##  ##     ##
##     ## ########
######### ##
##     ## ##
##     ## ##
*/

void activate_AP()
{
  factory_mode = true;
  // configure AP mode
  // activate watchdog for 3 minutes

  Serial.begin(115200);
  Serial.println();
  Serial.println("Factory reset ...");
  if (!SPIFFS.begin(true))
  {
    Serial.println("An Error has occurred while mounting SPIFFS");
    return;
  }

  // You can remove the password parameter if you want the AP to be open.

  WiFi.softAP(config.apssid, config.appass);
  IPAddress myIP = WiFi.softAPIP();
  Serial.print(F("AP IP address: "));
  Serial.println(myIP);
  Serial.println("Server started");
  WiFi.macAddress(MAC_array);
  for (int i = 0; i < sizeof(MAC_array); ++i)
  {
    sprintf(MAC_char, "%s%02x", MAC_char, MAC_array[i]);
  }
  Serial.println(MAC_char);
  startServer();
}

/*
##      ## ######## ########
##  ##  ## ##       ##     ##
##  ##  ## ##       ##     ##
##  ##  ## ######   ########
##  ##  ## ##       ##     ##
##  ##  ## ##       ##     ##
 ###  ###  ######## ########

*/

/**
 * @brief Start WEB server
 *
 */
void startServer()
{
  setSound(200);
  // Route for root / web page
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request)
            { request->send(SPIFFS, "/index.html", String(), false, processor); });

  // Route to load style.css file
  server.on("/style.css", HTTP_GET, [](AsyncWebServerRequest *request)
            { request->send(SPIFFS, "/style.css", "text/css"); });

  // Send a POST request to <IP>/post with a form field message set to <message>
  server.on("/config", HTTP_POST, [](AsyncWebServerRequest *request)
            {
              String message = "";
              String param;
              config.dhcp=0;
              config.sound=0;
              if (request->hasParam(PARAM_NAME, true))
              {
                param = request->getParam(PARAM_NAME, true)->value();
                message += "NAME:";
                message += param;
                message += "\r\n";
                strlcpy(config.name, param.c_str(), sizeof(config.name));
              }
              if (request->hasParam(PARAM_SSID, true))
              {
                param = request->getParam(PARAM_SSID, true)->value();
                message += "SSID:";
                message += param;
                message += "\r\n";
                strlcpy(config.ssid, param.c_str(), sizeof(config.ssid));
              }
              if (request->hasParam(PARAM_PASS, true))
              {
                param = request->getParam(PARAM_PASS, true)->value();
                message += "PASS:";
                message += param;
                message += "\r\n";
                strlcpy(config.pass, param.c_str(), sizeof(config.pass));
              }
              if (request->hasParam(PARAM_MQTT, true))
              {
                param = request->getParam(PARAM_MQTT, true)->value();
                message += "MQTT:";
                message += param;
                message += "\r\n";
                strlcpy(config.mqtt, param.c_str(), sizeof(config.mqtt));
              }
              if (request->hasParam(PARAM_MQTT_PORT, true))
              {
                param = request->getParam(PARAM_MQTT_PORT, true)->value();
                message += "MQTT_PORT:";
                message += param;
                message += "\r\n";
                config.mqtt_port = tools.StringTo64(param);
              }
               if (request->hasParam(PARAM_MQTT_USER, true))
              {
                param = request->getParam(PARAM_MQTT_USER, true)->value();
                message += "MQTT_USER:";
                message += param;
                message += "\r\n";
                strlcpy(config.mqtt_user, param.c_str(), sizeof(config.mqtt_user));
              }
              if (request->hasParam(PARAM_MQTT_USER_PASS, true))
              {
                param = request->getParam(PARAM_MQTT_USER_PASS, true)->value();
                message += "MQTT_USER_PASS:";
                message += param;
                message += "\r\n";
                strlcpy(config.mqtt_user_pass, param.c_str(), sizeof(config.mqtt_user_pass));
              }
              if (request->hasParam(PARAM_DHCP, true))
              {
                param = request->getParam(PARAM_DHCP, true)->value();
                message += "DHCP:";
                message += param;
                message += "\r\n";
                config.dhcp = 1;
              }              
              if (request->hasParam(PARAM_IP, true))
              {
                param = request->getParam(PARAM_IP, true)->value();
                message += "IP:";
                message += param;
                message += "\r\n";
                strlcpy(config.ip_address, param.c_str(), sizeof(config.ip_address));
              }
              if (request->hasParam(PARAM_MASK, true))
              {
                param = request->getParam(PARAM_MASK, true)->value();
                message += "MASK:";
                message += param;
                message += "\r\n";
                strlcpy(config.ip_mask, param.c_str(), sizeof(config.ip_mask));
              }
              if (request->hasParam(PARAM_GATEWAY, true))
              {
                param = request->getParam(PARAM_GATEWAY, true)->value();
                message += "GATEWAY:";
                message += param;
                message += "\r\n";
                strlcpy(config.gateway, param.c_str(), sizeof(config.gateway));
              }
              if (request->hasParam(PARAM_TYPE, true))
              {
                param = request->getParam(PARAM_TYPE, true)->value();
                message += "TYPE:";
                message += param;
                message += "\r\n";
                strlcpy(config.type, param.c_str(), sizeof(config.type));
              }
               if (request->hasParam(PARAM_GROUP, true))
              {
                param = request->getParam(PARAM_GROUP, true)->value();
                message += "GROUP:";
                message += param;
                message += "\r\n";
                config.group = (uint16_t) tools.StringTo64(param);
              }
              if (request->hasParam(PARAM_HOTEL, true))
              {
                param = request->getParam(PARAM_HOTEL, true)->value();
                message += "HOTEL:";
                message += param;
                message += "\r\n";
                config.hotel_code = tools.StringTo64(param);
              }
              if (request->hasParam(PARAM_SOUND, true))
              {
                param = request->getParam(PARAM_SOUND, true)->value();
                message += "SOUND:";
                message += param;
                message += "\r\n";
                config.sound = 1;
              }
              
              Serial << message << endl;
              request->send(200, "text/plain", message);

              saveConfiguration(); 
              restart(); });

  server.on("/exit", HTTP_GET, [](AsyncWebServerRequest *request)
            { restart(); });

  // server.on("/update", HTTP_POST, [](AsyncWebServerRequest *request)
  //           { update(request); });

  // MDNS.begin(host);
  // MDNS.addService("http", "tcp", 80);

  // Start server
  AsyncElegantOTA.begin(&server);
  server.begin();
}

/**
 * @brief Updates state in html page
 *
 * @param var
 * @return String
 */
String processor(const String &var)
{
  Serial.println(var);
  if (var == "ap_ssid")
  {
    return config.apssid;
  }
  else if (var == "ssid")
  {
    return config.ssid;
  }
  else if (var == "pass")
  {
    return config.pass;
  }
  else if (var == "ip_address")
  {
    return config.ip_address;
  }
  else if (var == "event")
  {
    return config.event;
  }
  else if (var == "ip_mask")
  {
    return config.ip_mask;
  }
  else if (var == "name")
  {
    return config.name;
  }
  else if (var == "relay_time")
  {
    return String(config.relay_time, DEC);
  }
  else if (var == "mqtt")
  {
    return config.mqtt;
  }
  else if (var == "mqtt_port")
  {
    return String(config.mqtt_port, DEC);
  }
  else if (var == "sound")
  {
    if (config.sound == 1)
    {
      return "true";
    }
    return "false";
  }
  else if (var == "dhcp")
  {
    if (config.dhcp == 1)
    {
      return "true";
    }
    return "false";
  }
  return String();
}

/**
 * @brief Saves the configuration to a file
 *
 * @param filename
 * @param config
 */
void saveConfiguration()
{
  // Delete existing file, otherwise the configuration is appended to the file
  SPIFFS.remove(CONNFIG_FILE);

  // Open file for writing
  File file = SPIFFS.open(CONNFIG_FILE, FILE_WRITE);
  if (!file)
  {
    Serial.println(F("Failed to create file"));
    return;
  }

  // Allocate a temporary JsonDocument
  // Don't forget to change the capacity to match your requirements.
  // Use arduinojson.org/assistant to compute the capacity.
  StaticJsonDocument<1024> doc;

  // Set the values in the document
  doc["dhcp"] = config.dhcp;
  doc["event"] = config.event;
  doc["ip_address"] = config.ip_address;
  doc["ip_mask"] = config.ip_mask;
  doc["mqtt"] = config.mqtt;
  doc["mqtt_port"] = config.mqtt_port;
  doc["pass"] = config.pass;
  doc["relay_time"] = config.relay_time;
  doc["sound"] = config.sound;
  doc["ssid"] = config.ssid;
  doc["type"] = config.type;
  doc["apssid"] = config.apssid;
  doc["appass"] = config.appass;
  doc["name"] = config.name;
  doc["device_mac"] = MAC_char;
  doc["mqtt_user"] = config.mqtt_user;
  doc["mqtt_user_pass"] = config.mqtt_user_pass;
  doc["hotel_code"] = config.hotel_code;
  doc["gateway"] = config.gateway;
  doc["room"] = config.room;
  doc["group"] = config.group;
  doc["update_code"] = config.update_code;

  // Serialize JSON to file
  if (serializeJson(doc, file) == 0)
  {
    Serial.println(F("Failed to write to file"));
  }
  else
  {
    Serial.println("Config file saved");
  }
  // Close the file
  file.close();
  filesys.printFile(SPIFFS, CONNFIG_FILE);
  serializeJson(doc, Serial);
}

void restart()
{
  esp_sleep_enable_timer_wakeup(1000000); // sleep 1s
  Serial.println(F("Restarting..."));
  esp_deep_sleep_start();
}

// ##        #######     ###    ########      ######   #######  ##    ## ######## ####  ######
// ##       ##     ##   ## ##   ##     ##    ##    ## ##     ## ###   ## ##        ##  ##    ##
// ##       ##     ##  ##   ##  ##     ##    ##       ##     ## ####  ## ##        ##  ##
// ##       ##     ## ##     ## ##     ##    ##       ##     ## ## ## ## ######    ##  ##   ####
// ##       ##     ## ######### ##     ##    ##       ##     ## ##  #### ##        ##  ##    ##
// ##       ##     ## ##     ## ##     ##    ##    ## ##     ## ##   ### ##        ##  ##    ##  ###
// ########  #######  ##     ## ########      ######   #######  ##    ## ##       ####  ######   ###

/**
 * @brief Load configuration file
 *
 */
void loadConfiguration()
{
  Serial.println("Reading config file");

  File file = SPIFFS.open(CONNFIG_FILE);
  if (!file)
  {
    Serial.println("Failed to open file for reading");
    return;
  }
  Serial.println("Config file loaded");
  StaticJsonDocument<1024> doc;
  // Deserialize the JSON document
  DeserializationError error = deserializeJson(doc, file);
  file.close();
  if (error == error.IncompleteInput)
  {
    Serial << "incomplete" << endl;
  }
  if (error == error.InvalidInput)
  {
    Serial << "invaliv input" << endl;
  }
  if (error == error.EmptyInput)
  {
    Serial << "empty input" << endl;
  }
  if (error == error.NoMemory)
  {
    Serial << "no memory" << endl;
  }
  if (error == error.TooDeep)
  {
    Serial << "too deep" << endl;
  }
  if (error == error.Ok)
  {
    Serial << "OK" << endl;
  }

  if (error)
  {
    Serial.println(F("Deserialization Error"));
    setSound(150);
    return;
  }

  // Copy values from the JsonDocument to the Config
  config.dhcp = doc["dhcp"] | 1;
  strlcpy(config.event,                                                                           // <- destination
          doc["event"] | "access",                                                                // <- source
          sizeof(config.event));                                                                  // <- destination's capacity
  strlcpy(config.ip_address, doc["ip_address"] | "192.168.1.100", sizeof(config.ip_address) + 1); // AP
  strlcpy(config.ip_mask, doc["ip_mask"] | "255.255.255.0", sizeof(config.ip_mask) + 1);
  strlcpy(config.mqtt, doc["mqtt"] | "192.168.1.15", sizeof(config.mqtt) + 1);
  config.mqtt_port = doc["mqtt_port"] | 1883;
  strlcpy(config.pass, doc["pass"] | "4ETj5dPy", sizeof(config.pass) + 1);
  config.relay_time = doc["relay_time"] | 3;
  config.sound = doc["sound"] | 1;
  strlcpy(config.ssid, doc["ssid"] | "biodit_sys", sizeof(config.ssid) + 1);
  strlcpy(config.apssid, doc["apssid"] | "BioditWriter", sizeof(config.apssid) + 1);
  strlcpy(config.type, doc["type"] | "writer", sizeof(config.type) + 1);
  strlcpy(config.name, doc["name"] | "Room", sizeof(config.name) + 1);
  strlcpy(config.appass, doc["appass"] | "biodit1234", sizeof(config.pass) + 1);
  strlcpy(config.mqtt_user, doc["mqtt_user"] | "olock", sizeof(config.mqtt_user) + 1);
  strlcpy(config.mqtt_user_pass, doc["mqtt_user_pass"] | "pass", sizeof(config.mqtt_user_pass) + 1);
  config.hotel_code = doc["hotel_code"] | 1;
  config.room = doc["room"] | 1;
  config.group = doc["group"] | 1;
  strlcpy(config.update_code, doc["update_code"] | "t3o6z7fryx3lrt1c", sizeof(config.update_code) + 1);
}

void blink()
{
  setSound(100);
  digitalWrite(LED_RED, 1);
  delay(100);
  digitalWrite(LED_RED, 0);
}